"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const compile_1 = require("./compile");
const util_1 = require("./utils/util");
const blockchain_1 = require("./blockchain");
const downloadCompiler_1 = require("./downloadCompiler");
const fs = require('fs');
class Vscc {
    constructor(rpcEndpoint, _path) {
        if (!fs.statSync(_path).isDirectory()) {
            throw Error('Init Error. BinsPath invalid');
        }
        this.binPath = _path;
        this.thor_node_rpc_endpoint = rpcEndpoint;
        this.blockchain = new blockchain_1.BlockChain(this.thor_node_rpc_endpoint);
    }
    /**
     * verify contract
     * @param params params verify contract params
     */
    verify(params) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!util_1.verifyAddress(params.contractAddress)) {
                throw Error('contract address invalid.');
            }
            if (!params.sourceCode) {
                throw Error('sourceCode invalid');
            }
            if (!params.contractName) {
                throw Error('contract_name invalid');
            }
            if (!util_1.verifyTxId(params.txId)) {
                throw Error('txId invalid');
            }
            if (params.solcVersion) {
                const vPath = util_1.solcBinPath(params.solcVersion, this.binPath);
                if (!fs.existsSync(vPath)) {
                    const versionList = yield downloadCompiler_1.getVersionList();
                    const versionName = versionList.releases[params.solcVersion];
                    if (versionName) {
                        throw Error(`please download compiler version first. [${params.solcVersion}]`);
                    }
                    else {
                        throw Error(`version [${params.solcVersion}] invalid`);
                    }
                }
            }
            // verify contract address
            const hasCode = yield this.verifyContractAddress(params.contractAddress);
            if (!hasCode) {
                throw Error('contract address invalid');
            }
            // deploy code
            const deployedCode = yield this.getCode(params.contractAddress);
            // solc compile code
            const compileCode = yield compile_1.solcCompile(util_1.decodeBase64(params.sourceCode), params.contractName, params.solcVersion, parseInt(params.runs || '200'), this.binPath, params.isOptimizer);
            if (compileCode.length === 0) {
                throw Error('compile failed');
            }
            const _params = params.params ? params.params : '';
            if (_params.length % 64 !== 0) {
                throw Error('constructor params invalid');
            }
            // deploy contract call
            const deployCode = yield this.deployContractOfCall(compileCode).catch(err => {
                throw Error(`deploy contract call error. ${err}`);
            });
            // compare runtimeBytecode
            const compare = this.compareByteCode(deployedCode, deployCode);
            if (!compare) {
                throw Error('verify contract source code failed');
            }
            // verify contract encodeAbi
            yield this.matchContractParamsAbiCode(params.txId, compileCode, params.params).catch(err => {
                throw err;
            });
            return true;
        });
    }
    /**
     * compare runtime bytecode
     * @param deployedCode deployed contract runtime bytecode
     * @param deployCode deploy contract runtime bytecode of call
     */
    compareByteCode(deployedCode, deployCode) {
        const deployedCodeSubStr = deployedCode.substring(0, deployedCode.length - 68);
        const deployCodeSubStr = deployCode.substring(0, deployCode.length - 68);
        return deployedCodeSubStr === deployCodeSubStr;
    }
    /**
     * get code from address
     * @param contract_address contract address
     */
    getCode(contract_address) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.blockchain.getCode(contract_address).catch(err => {
                throw 'Get code error: ' + err.toString();
            });
            return result.code;
        });
    }
    /**
     * call deploy contract
     * @param byteCode byteCode for deploy contract
     */
    deployContractOfCall(byteCode) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.blockchain.deployContractOfCall(byteCode).catch(err => {
                throw 'Deploy contract call error. ' + err.toString();
            });
            return result.data;
        });
    }
    /**
     * verify address whether contract
     * @param address contract address
     */
    verifyContractAddress(address) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.blockchain.accounts(address).catch(err => {
                throw err.toString();
            });
            return result.hasCode;
        });
    }
    /**
     * verify contract constructor encode abi
     * @param txId contract deploy transaction id
     * @param params contract deploy encode abi
     */
    matchContractParamsAbiCode(txId, compileCode, params = '') {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.blockchain.getTransaction(txId).catch(err => {
                throw err.toString();
            });
            if (!result) {
                throw 'txId invalid';
            }
            // TODO: only one clause
            result.clauses.forEach(clause => {
                const data = clause.data;
                if ((data.length - 2) === (compileCode.length + params.length)) {
                    const chainCodeParams = data.substr(compileCode.length + 2);
                    if (chainCodeParams !== params) {
                        throw Error('verify contract source code failed');
                    }
                }
                else {
                    throw Error('verify contract source code failed');
                }
            });
        });
    }
}
exports.Vscc = Vscc;
//# sourceMappingURL=vscc.js.map